from __future__ import absolute_import, unicode_literals

__all__ = ['Graph']

from .api import APIWrapper
from .collection import EdgeCollection
from .collection import VertexCollection
from .exceptions import (
    EdgeDefinitionListError,
    EdgeDefinitionCreateError,
    EdgeDefinitionDeleteError,
    EdgeDefinitionReplaceError,
    GraphPropertiesError,
    GraphTraverseError,
    VertexCollectionListError,
    VertexCollectionCreateError,
    VertexCollectionDeleteError,
)
from .request import Request
from .utils import get_col_name, get_doc_id


class Graph(APIWrapper):
    """Graph API wrapper.

    :param executor: API executor.
    :type executor: arango.executor.Executor
    :param name: Graph name.
    :type name: str | unicode
    """

    def __init__(self, connection, executor, name):
        super(Graph, self).__init__(connection, executor)
        self._name = name

    def __repr__(self):
        return '<Graph {}>'.format(self._name)

    def _get_col_by_vertex(self, vertex):
        """Return the vertex collection for the given vertex document.

        :param vertex: Vertex document ID or body with "_id" field.
        :type vertex: str | unicode | dict
        :return: Vertex collection API wrapper.
        :rtype: arango.collection.VertexCollection
        """
        return self.vertex_collection(get_col_name(vertex))

    def _get_col_by_edge(self, edge):
        """Return the vertex collection for the given edge document.

        :param edge: Edge document ID or body with "_id" field.
        :type edge: str | unicode | dict
        :return: Edge collection API wrapper.
        :rtype: arango.collection.EdgeCollection
        """
        return self.edge_collection(get_col_name(edge))

    @property
    def name(self):
        """Return the graph name.

        :return: Graph name.
        :rtype: str | unicode
        """
        return self._name

    async def properties(self):
        """Return graph properties.

        :return: Graph properties.
        :rtype: dict
        :raise arango.exceptions.GraphPropertiesError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/gharial/{}'.format(self._name)
        )

        def response_handler(resp):  # pragma: no cover
            if not resp.is_success:
                raise GraphPropertiesError(resp, request)
            body = resp.body['graph']
            properties = {
                'id': body['_id'],
                'name': body['name'],
                'revision': body['_rev'],
                'orphan_collections': body['orphanCollections'],
                'edge_definitions': [
                    {
                        'edge_collection': edge_definition['collection'],
                        'from_vertex_collections': edge_definition['from'],
                        'to_vertex_collections': edge_definition['to'],
                    }
                    for edge_definition in body['edgeDefinitions']
                ]
            }
            if 'isSmart' in body:
                properties['smart'] = body['isSmart']
            if 'smartGraphAttribute' in body:
                properties['smart_field'] = body['smartGraphAttribute']
            if 'numberOfShards' in body:
                properties['shard_count'] = body['numberOfShards']
            if 'replicationFactor' in body:
                properties['replication_factor'] = body['replicationFactor']
            return properties

        return await self._execute(request, response_handler)

    ################################
    # Vertex Collection Management #
    ################################

    async def has_vertex_collection(self, name):
        """Check if the graph has the given vertex collection.

        :param name: Vertex collection name.
        :type name: str | unicode
        :return: True if vertex collection exists, False otherwise.
        :rtype: bool
        """
        return name in await self.vertex_collections()

    async def vertex_collections(self):
        """Return vertex collections in the graph that are not orphaned.

        :return: Names of vertex collections that are not orphaned.
        :rtype: [str | unicode]
        :raise arango.exceptions.VertexCollectionListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/gharial/{}/vertex'.format(self._name),
        )

        def response_handler(resp):
            if not resp.is_success:
                raise VertexCollectionListError(resp, request)
            return sorted(set(resp.body['collections']))

        return await self._execute(request, response_handler)

    def vertex_collection(self, name):
        """Return the vertex collection API wrapper.

        :param name: Vertex collection name.
        :type name: str | unicode
        :return: Vertex collection API wrapper.
        :rtype: arango.collection.VertexCollection
        """
        return VertexCollection(self._conn, self._executor, self._name, name)

    async def create_vertex_collection(self, name):
        """Create a vertex collection in the graph.

        :param name: Vertex collection name.
        :type name: str | unicode
        :return: Vertex collection API wrapper.
        :rtype: arango.collection.VertexCollection
        :raise arango.exceptions.VertexCollectionCreateError: If create fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/gharial/{}/vertex'.format(self._name),
            data={'collection': name}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise VertexCollectionCreateError(resp, request)
            return self.vertex_collection(name)

        return await self._execute(request, response_handler)

    async def delete_vertex_collection(self, name, purge=False):
        """Remove a vertex collection from the graph.

        :param name: Vertex collection name.
        :type name: str | unicode
        :param purge: If set to True, the vertex collection is not just deleted
            from the graph but also from the database completely.
        :type purge: bool
        :return: True if vertex collection was deleted successfully.
        :rtype: bool
        :raise arango.exceptions.VertexCollectionDeleteError: If delete fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/gharial/{}/vertex/{}'.format(self._name, name),
            params={'dropCollection': purge}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise VertexCollectionDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    ##############################
    # Edge Collection Management #
    ##############################

    async def has_edge_definition(self, name):
        """Check if the graph has the given edge definition.

        :param name: Edge collection name.
        :type name: str | unicode
        :return: True if edge definition exists, False otherwise.
        :rtype: bool
        """
        return any(
            definition['edge_collection'] == name
            for definition in await self.edge_definitions()
        )

    async def has_edge_collection(self, name):
        """Check if the graph has the given edge collection.

        :param name: Edge collection name.
        :type name: str | unicode
        :return: True if edge collection exists, False otherwise.
        :rtype: bool
        """
        return await self.has_edge_definition(name)

    def edge_collection(self, name):
        """Return the edge collection API wrapper.

        :param name: Edge collection name.
        :type name: str | unicode
        :return: Edge collection API wrapper.
        :rtype: arango.collection.EdgeCollection
        """
        return EdgeCollection(self._conn, self._executor, self._name, name)

    async def edge_definitions(self):
        """Return the edge definitions of the graph.

        :return: Edge definitions of the graph.
        :rtype: [dict]
        :raise arango.exceptions.EdgeDefinitionListError: If retrieval fails.
        """
        try:
            return (await self.properties())['edge_definitions']
        except GraphPropertiesError as err:
            raise EdgeDefinitionListError(err.response, err.request)

    async def create_edge_definition(
            self,
            edge_collection,
            from_vertex_collections,
            to_vertex_collections):
        """Create a new edge definition.

        An edge definition consists of an edge collection, "from" vertex
        collection(s) and "to" vertex collection(s). Here is an example entry:

        .. code-block:: python

            {
                'edge_collection': 'edge_collection_name',
                'from_vertex_collections': ['from_vertex_collection_name'],
                'to_vertex_collections': ['to_vertex_collection_name']
            }

        :param edge_collection: Edge collection name.
        :type edge_collection: str | unicode
        :param from_vertex_collections: Names of "from" vertex collections.
        :type from_vertex_collections: [str | unicode]
        :param to_vertex_collections: Names of "to" vertex collections.
        :type to_vertex_collections: [str | unicode]
        :return: Edge collection API wrapper.
        :rtype: arango.collection.EdgeCollection
        :raise arango.exceptions.EdgeDefinitionCreateError: If create fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/gharial/{}/edge'.format(self._name),
            data={
                'collection': edge_collection,
                'from': from_vertex_collections,
                'to': to_vertex_collections
            }
        )

        def response_handler(resp):
            if not resp.is_success:
                raise EdgeDefinitionCreateError(resp, request)
            return self.edge_collection(edge_collection)

        return await self._execute(request, response_handler)

    async def replace_edge_definition(self,
                                edge_collection,
                                from_vertex_collections,
                                to_vertex_collections):
        """Replace an edge definition.

        :param edge_collection: Edge collection name.
        :type edge_collection: str | unicode
        :param from_vertex_collections: Names of "from" vertex collections.
        :type from_vertex_collections: [str | unicode]
        :param to_vertex_collections: Names of "to" vertex collections.
        :type to_vertex_collections: [str | unicode]
        :return: True if edge definition was replaced successfully.
        :rtype: bool
        :raise arango.exceptions.EdgeDefinitionReplaceError: If replace fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/gharial/{}/edge/{}'.format(
                self._name, edge_collection
            ),
            data={
                'collection': edge_collection,
                'from': from_vertex_collections,
                'to': to_vertex_collections
            }
        )

        def response_handler(resp):
            if not resp.is_success:
                raise EdgeDefinitionReplaceError(resp, request)
            return self.edge_collection(edge_collection)

        return await self._execute(request, response_handler)

    async def delete_edge_definition(self, name, purge=False):
        """Delete an edge definition from the graph.

        :param name: Edge collection name.
        :type name: str | unicode
        :param purge: If set to True, the edge definition is not just removed
            from the graph but the edge collection is also deleted completely
            from the database.
        :type purge: bool
        :return: True if edge definition was deleted successfully.
        :rtype: bool
        :raise arango.exceptions.EdgeDefinitionDeleteError: If delete fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/gharial/{}/edge/{}'.format(self._name, name),
            params={'dropCollections': purge}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise EdgeDefinitionDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    ###################
    # Graph Functions #
    ###################

    async def traverse(
            self,
            start_vertex,
            direction='outbound',
            item_order='forward',
            strategy=None,
            order=None,
            edge_uniqueness=None,
            vertex_uniqueness=None,
            max_iter=None,
            min_depth=None,
            max_depth=None,
            init_func=None,
            sort_func=None,
            filter_func=None,
            visitor_func=None,
            expander_func=None):
        """Traverse the graph and return the visited vertices and edges.

        :param start_vertex: Start vertex document ID or body with "_id" field.
        :type start_vertex: str | unicode | dict
        :param direction: Traversal direction. Allowed values are "outbound"
            (default), "inbound" and "any".
        :type direction: str | unicode
        :param item_order: Item iteration order. Allowed values are "forward"
            (default) and "backward".
        :type item_order: str | unicode
        :param strategy: Traversal strategy. Allowed values are "depthfirst"
            and "breadthfirst".
        :type strategy: str | unicode
        :param order: Traversal order. Allowed values are "preorder",
            "postorder", and "preorder-expander".
        :type order: str | unicode
        :param vertex_uniqueness: Uniqueness for visited vertices. Allowed
            values are "global", "path" or "none".
        :type vertex_uniqueness: str | unicode
        :param edge_uniqueness: Uniqueness for visited edges. Allowed values
            are "global", "path" or "none".
        :type edge_uniqueness: str | unicode
        :param min_depth: Minimum depth of the nodes to visit.
        :type min_depth: int
        :param max_depth: Maximum depth of the nodes to visit.
        :type max_depth: int
        :param max_iter: If set, halt the traversal after the given number of
            iterations. This parameter can be used to prevent endless loops in
            cyclic graphs.
        :type max_iter: int
        :param init_func: Initialization function in Javascript with signature
            ``(config, result) -> void``. This function is used to initialize
            values in the result.
        :type init_func: str | unicode
        :param sort_func: Sorting function in Javascript with signature
            ``(left, right) -> integer``, which returns ``-1`` if ``left <
            right``, ``+1`` if ``left > right`` and ``0`` if ``left == right``.
        :type sort_func: str | unicode
        :param filter_func: Filter function in Javascript with signature
            ``(config, vertex, path) -> mixed``, where ``mixed`` can have one
            of the following values (or an array with multiple): "exclude" (do
            not visit the vertex), "prune" (do not follow the edges of the
            vertex), or "undefined" (visit the vertex and follow its edges).
        :type filter_func: str | unicode
        :param visitor_func: Visitor function in Javascript with signature
            ``(config, result, vertex, path, connected) -> void``. The return
            value is ignored, ``result`` is modified by reference, and
            ``connected`` is populated only when parameter **order** is set to
            "preorder-expander".
        :type visitor_func: str | unicode
        :param expander_func: Expander function in Javascript with signature
            ``(config, vertex, path) -> mixed``. The function must return an
            array of connections for ``vertex``. Each connection is an object
            with attributes "edge" and "vertex".
        :type expander_func: str | unicode
        :return: Visited edges and vertices.
        :rtype: dict
        :raise arango.exceptions.GraphTraverseError: If traversal fails.
        """
        if strategy is not None:
            if strategy.lower() == 'dfs':
                strategy = 'depthfirst'
            elif strategy.lower() == 'bfs':
                strategy = 'breadthfirst'

        uniqueness = {}
        if vertex_uniqueness is not None:
            uniqueness['vertices'] = vertex_uniqueness
        if edge_uniqueness is not None:
            uniqueness['edges'] = edge_uniqueness

        data = {
            'startVertex': get_doc_id(start_vertex),
            'graphName': self._name,
            'direction': direction,
            'strategy': strategy,
            'order': order,
            'itemOrder': item_order,
            'uniqueness': uniqueness or None,
            'maxIterations': max_iter,
            'minDepth': min_depth,
            'maxDepth': max_depth,
            'init': init_func,
            'filter': filter_func,
            'visitor': visitor_func,
            'sort': sort_func,
            'expander': expander_func
        }
        request = Request(
            method='post',
            endpoint='/_api/traversal',
            data={k: v for k, v in data.items() if v is not None}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise GraphTraverseError(resp, request)
            return resp.body['result']['visited']

        return await self._execute(request, response_handler)

    #####################
    # Vertex Management #
    #####################

    async def has_vertex(self, vertex, rev=None, check_rev=True):
        """Check if the given vertex document exists in the graph.

        :param vertex: Vertex document ID or body with "_id" field.
        :type vertex: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **vertex** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :return: True if vertex document exists, False otherwise.
        :rtype: bool
        :raise arango.exceptions.DocumentGetError: If check fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_vertex(vertex).has(vertex, rev, check_rev)

    async def vertex(self, vertex, rev=None, check_rev=True):
        """Return a vertex document.

        :param vertex: Vertex document ID or body with "_id" field.
        :type vertex: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **vertex** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :return: Vertex document or None if not found.
        :rtype: dict | None
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_vertex(vertex).get(vertex, rev, check_rev)

    async def insert_vertex(self, collection, vertex, sync=None, silent=False):
        """Insert a new vertex document.

        :param collection: Vertex collection name.
        :type collection: str | unicode
        :param vertex: New vertex document to insert. If it has "_key" or "_id"
            field, its value is used as key of the new vertex (otherwise it is
            auto-generated). Any "_rev" field is ignored.
        :type vertex: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        return await self.vertex_collection(collection).insert(vertex, sync, silent)

    async def update_vertex(
            self,
            vertex,
            check_rev=True,
            keep_none=True,
            sync=None,
            silent=False):
        """Update a vertex document.

        :param vertex: Partial or full vertex document with updated values. It
            must contain the "_id" field.
        :type vertex: dict
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. If set to False, they are removed completely.
        :type keep_none: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_vertex(vertex).update(
            vertex=vertex,
            check_rev=check_rev,
            keep_none=keep_none,
            sync=sync,
            silent=silent
        )

    async def replace_vertex(self, vertex, check_rev=True, sync=None, silent=False):
        """Replace a vertex document.

        :param vertex: New vertex document to replace the old one with. It must
            contain the "_id" field.
        :type vertex: dict
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_vertex(vertex).replace(
            vertex=vertex,
            check_rev=check_rev,
            sync=sync,
            silent=silent
        )

    async def delete_vertex(
            self,
            vertex,
            rev=None,
            check_rev=True,
            ignore_missing=False,
            sync=None):
        """Delete a vertex document.

        :param vertex: Vertex document ID or body with "_id" field.
        :type vertex: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **vertex** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **vertex** (if given) is
            compared against the revision of target vertex document.
        :type check_rev: bool
        :param ignore_missing: Do not raise an exception on missing document.
            This parameter has no effect in transactions where an exception is
            always raised on failures.
        :type ignore_missing: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :return: True if vertex was deleted successfully, False if vertex was
            not found and **ignore_missing** was set to True (does not apply in
            transactions).
        :rtype: bool
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_vertex(vertex).delete(
            vertex=vertex,
            rev=rev,
            check_rev=check_rev,
            ignore_missing=ignore_missing,
            sync=sync
        )

    ###################
    # Edge Management #
    ###################

    async def has_edge(self, edge, rev=None, check_rev=True):
        """Check if the given edge document exists in the graph.

        :param edge: Edge document ID or body with "_id" field.
        :type edge: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **edge** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :return: True if edge document exists, False otherwise.
        :rtype: bool
        :raise arango.exceptions.DocumentInError: If check fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_edge(edge).has(edge, rev, check_rev)

    async def edge(self, edge, rev=None, check_rev=True):
        """Return an edge document.

        :param edge: Edge document ID or body with "_id" field.
        :type edge: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **edge** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :return: Edge document or None if not found.
        :rtype: dict | None
        :raise arango.exceptions.DocumentGetError: If retrieval fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_edge(edge).get(edge, rev, check_rev)

    async def insert_edge(self, collection, edge, sync=None, silent=False):
        """Insert a new edge document.

        :param collection: Edge collection name.
        :type collection: str | unicode
        :param edge: New edge document to insert. It must contain "_from" and
            "_to" fields. If it has "_key" or "_id" field, its value is used
            as key of the new edge document (otherwise it is auto-generated).
            Any "_rev" field is ignored.
        :type edge: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        return await self.edge_collection(collection).insert(edge, sync, silent)

    async def update_edge(
            self,
            edge,
            check_rev=True,
            keep_none=True,
            sync=None,
            silent=False):
        """Update an edge document.

        :param edge: Partial or full edge document with updated values. It must
            contain the "_id" field.
        :type edge: dict
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :param keep_none: If set to True, fields with value None are retained
            in the document. If set to False, they are removed completely.
        :type keep_none: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentUpdateError: If update fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_edge(edge).update(
            edge=edge,
            check_rev=check_rev,
            keep_none=keep_none,
            sync=sync,
            silent=silent
        )

    async def replace_edge(self, edge, check_rev=True, sync=None, silent=False):
        """Replace an edge document.

        :param edge: New edge document to replace the old one with. It must
            contain the "_id" field. It must also contain the "_from" and "_to"
            fields.
        :type edge: dict
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentReplaceError: If replace fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_edge(edge).replace(
            edge=edge,
            check_rev=check_rev,
            sync=sync,
            silent=silent
        )

    async def delete_edge(
            self,
            edge,
            rev=None,
            check_rev=True,
            ignore_missing=False,
            sync=None):
        """Delete an edge document.

        :param edge: Edge document ID or body with "_id" field.
        :type edge: str | unicode | dict
        :param rev: Expected document revision. Overrides the value of "_rev"
            field in **edge** if present.
        :type rev: str | unicode
        :param check_rev: If set to True, revision of **edge** (if given) is
            compared against the revision of target edge document.
        :type check_rev: bool
        :param ignore_missing: Do not raise an exception on missing document.
            This parameter has no effect in transactions where an exception is
            always raised on failures.
        :type ignore_missing: bool
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :return: True if edge was deleted successfully, False if edge was not
            found and **ignore_missing** was set to True (does not  apply in
            transactions).
        :rtype: bool
        :raise arango.exceptions.DocumentDeleteError: If delete fails.
        :raise arango.exceptions.DocumentRevisionError: If revisions mismatch.
        """
        return await self._get_col_by_edge(edge).delete(
            edge=edge,
            rev=rev,
            check_rev=check_rev,
            ignore_missing=ignore_missing,
            sync=sync
        )

    async def link(
            self,
            collection,
            from_vertex,
            to_vertex,
            data=None,
            sync=None,
            silent=False):
        """Insert a new edge document linking the given vertices.

        :param collection: Edge collection name.
        :type collection: str | unicode
        :param from_vertex: "From" vertex document ID or body with "_id" field.
        :type from_vertex: str | unicode | dict
        :param to_vertex: "To" vertex document ID or body with "_id" field.
        :type to_vertex: str | unicode | dict
        :param data: Any extra data for the new edge document. If it has "_key"
            or "_id" field, its value is used as key of the new edge document
            (otherwise it is auto-generated).
        :type data: dict
        :param sync: Block until operation is synchronized to disk.
        :type sync: bool
        :param silent: If set to True, no document metadata is returned. This
            can be used to save resources.
        :type silent: bool
        :return: Document metadata (e.g. document key, revision) or True if
            parameter **silent** was set to True.
        :rtype: bool | dict
        :raise arango.exceptions.DocumentInsertError: If insert fails.
        """
        return await self.edge_collection(collection).link(
            from_vertex=from_vertex,
            to_vertex=to_vertex,
            data=data,
            sync=sync,
            silent=silent
        )

    async def edges(self, collection, vertex, direction=None):
        """Return the edge documents coming in and/or out of given vertex.

        :param collection: Edge collection name.
        :type collection: str | unicode
        :param vertex: Vertex document ID or body with "_id" field.
        :type vertex: str | unicode | dict
        :param direction: The direction of the edges. Allowed values are "in"
            and "out". If not set, edges in both directions are returned.
        :type direction: str | unicode
        :return: List of edges and statistics.
        :rtype: dict
        :raise arango.exceptions.EdgeListError: If retrieval fails.
        """
        return await self.edge_collection(collection).edges(vertex, direction)



from __future__ import absolute_import, unicode_literals

import pytest
from aioarangodb.resolver import (
    SingleHostResolver,
    RandomHostResolver,
    RoundRobinHostResolver
)
pytestmark = pytest.mark.asyncio


async def test_resolver_single_host():
    resolver = SingleHostResolver()
    for _ in range(20):
        assert resolver.get_host_index() == 0


async def test_resolver_random_host():
    resolver = RandomHostResolver(10)
    for _ in range(20):
        assert 0 <= resolver.get_host_index() < 10


async def test_resolver_round_robin():
    resolver = RoundRobinHostResolver(10)
    assert resolver.get_host_index() == 0
    assert resolver.get_host_index() == 1
    assert resolver.get_host_index() == 2
    assert resolver.get_host_index() == 3
    assert resolver.get_host_index() == 4
    assert resolver.get_host_index() == 5
    assert resolver.get_host_index() == 6
    assert resolver.get_host_index() == 7
    assert resolver.get_host_index() == 8
    assert resolver.get_host_index() == 9
    assert resolver.get_host_index() == 0



from __future__ import absolute_import, unicode_literals

__all__ = ['Cluster']

from .api import APIWrapper
from .exceptions import (
    ClusterEndpointsError,
    ClusterHealthError,
    ClusterMaintenanceModeError,
    ClusterServerEngineError,
    ClusterServerIDError,
    ClusterServerRoleError,
    ClusterServerStatisticsError,
    ClusterServerVersionError,
)
from .request import Request


class Cluster(APIWrapper):  # pragma: no cover

    def __init__(self, connection, executor):
        super(Cluster, self).__init__(connection, executor)

    async def server_id(self):
        """Return the server ID.

        :return: Server ID.
        :rtype: str | unicode
        :raise arango.exceptions.ClusterServerIDError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/server/id'
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body['id']
            raise ClusterServerIDError(resp, request)

        return await self._execute(request, response_handler)

    async def server_role(self):
        """Return the server role.

        :return: Server role. Possible values are "SINGLE" (server which is
            not in a cluster), "COORDINATOR" (cluster coordinator), "PRIMARY",
            "SECONDARY", "AGENT" (Agency server in a cluster) or "UNDEFINED".
        :rtype: str | unicode
        :raise arango.exceptions.ClusterServerRoleError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/server/role'
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body['role']
            raise ClusterServerRoleError(resp, request)

        return await self._execute(request, response_handler)

    async def server_version(self, server_id):
        """Return the version of the given server.

        :param server_id: Server ID.
        :type server_id: str | unicode
        :return: Version of the given server.
        :rtype: dict
        :raise arango.exceptions.ClusterServerVersionError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/cluster/nodeVersion',
            params={'ServerID': server_id}
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise ClusterServerVersionError(resp, request)

        return await self._execute(request, response_handler)

    async def server_engine(self, server_id):
        """Return the engine details for the given server.

        :param server_id: Server ID.
        :type server_id: str | unicode
        :return: Engine details of the given server.
        :rtype: dict
        :raise arango.exceptions.ClusterServerEngineError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/cluster/nodeEngine',
            params={'ServerID': server_id}
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise ClusterServerEngineError(resp, request)

        return await self._execute(request, response_handler)

    async def server_statistics(self, server_id):
        """Return the statistics for the given server.

        :param server_id: Server ID.
        :type server_id: str | unicode
        :return: Statistics for the given server.
        :rtype: dict
        :raise arango.exceptions.ClusterServerStatisticsError: If retrieval
            fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/cluster/nodeStatistics',
            params={'ServerID': server_id}
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise ClusterServerStatisticsError(resp, request)

        return await self._execute(request, response_handler)

    async def health(self):
        """Return the cluster health.

        :return: Cluster health.
        :rtype: dict
        :raise arango.exceptions.ClusterHealthError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/cluster/health',
        )

        def response_handler(resp):
            if resp.is_success:
                resp.body.pop('error')
                resp.body.pop('code')
                return resp.body
            raise ClusterHealthError(resp, request)

        return await self._execute(request, response_handler)

    async def toggle_maintenance_mode(self, mode):
        """Enable or disable the cluster supervision (agency) maintenance mode.

        :param mode: Maintenance mode. Allowed values are "on" and "off".
        :type mode: str | unicode
        :return: Result of the operation.
        :rtype: dict
        :raise arango.exceptions.ClusterMaintenanceModeError: If toggle fails.
        """
        request = Request(
            method='put',
            endpoint='/_admin/cluster/maintenance',
            data='"{}"'.format(mode)
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise ClusterMaintenanceModeError(resp, request)

        return await self._execute(request, response_handler)

    async def endpoints(self):
        """Return coordinate endpoints. This method is for clusters only.

        :return: List of endpoints.
        :rtype: [str | unicode]
        :raise arango.exceptions.ServerEndpointsError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/cluster/endpoints'
        )

        def response_handler(resp):
            if not resp.is_success:
                raise ClusterEndpointsError(resp, request)
            return [item['endpoint'] for item in resp.body['endpoints']]

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

__all__ = ['Foxx']

import os

from aiohttp import MultipartWriter

from .api import APIWrapper
from .exceptions import (
    FoxxServiceCreateError,
    FoxxServiceDeleteError,
    FoxxServiceGetError,
    FoxxServiceListError,
    FoxxServiceReplaceError,
    FoxxServiceUpdateError,
    FoxxConfigGetError,
    FoxxConfigReplaceError,
    FoxxConfigUpdateError,
    FoxxDependencyGetError,
    FoxxDependencyReplaceError,
    FoxxDependencyUpdateError,
    FoxxScriptListError,
    FoxxScriptRunError,
    FoxxTestRunError,
    FoxxDevModeEnableError,
    FoxxDevModeDisableError,
    FoxxReadmeGetError,
    FoxxSwaggerGetError,
    FoxxCommitError,
    FoxxDownloadError,
)
from .request import Request


class Foxx(APIWrapper):
    """Foxx API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    """

    def __init__(self, connection, executor):
        super(Foxx, self).__init__(connection, executor)

    def __repr__(self):
        return '<Foxx in {}>'.format(self._conn.db_name)

    # noinspection PyMethodMayBeStatic
    def _encode_file(self, filename):
        """Encode file into multipart data.

        :param filename: Full path to the javascript file or zip bundle.
        :type filename: str | unicode
        :return: Multipart encoder object
        :rtype: aiohttp.MultipartWriter
        """
        extension = os.path.splitext(filename)[1]
        if extension == '.js':  # pragma: no cover
            source_type = 'application/javascript'
        elif extension == '.zip':
            source_type = 'application/zip'
        else:
            raise ValueError('File extension must be .zip or .js')

        mpwriter = MultipartWriter()
        mpwriter.append(
            open(filename, 'rb').read(),
            {'CONTENT-TYPE': source_type})
        return mpwriter

    async def services(self, exclude_system=False):
        """List installed services.

        :param exclude_system: If set to True, system services are excluded.
        :type exclude_system: bool
        :return: List of installed service.
        :rtype: [dict]
        :raise arango.exceptions.FoxxServiceListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx',
            params={'excludeSystem': exclude_system}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceListError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def service(self, mount):
        """Return service metadata.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx/service',
            params={'mount': mount}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceGetError(resp, request)

            if 'manifest' in resp.body:
                mf = resp.body['manifest']
                if 'defaultDocument' in mf:
                    mf['default_document'] = mf.pop('defaultDocument')

            return resp.body

        return await self._execute(request, response_handler)

    async def create_service(
            self,
            mount,
            source,
            config=None,
            dependencies=None,
            development=None,
            setup=None,
            legacy=None):
        """Install a new service using JSON definition.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param source: Fully qualified URL or absolute path on the server file
            system. Must be accessible by the server, or by all servers if in
            a cluster.
        :type source: str | unicode
        :param config: Configuration values.
        :type config: dict
        :param dependencies: Dependency settings.
        :type dependencies: dict
        :param development: Enable development mode.
        :type development: bool
        :param setup: Run service setup script.
        :type setup: bool
        :param legacy: Install the service in 2.8 legacy compatibility mode.
        :type legacy: bool
        :return: Service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceCreateError: If install fails.
        """
        params = {'mount': mount}
        if development is not None:
            params['development'] = development
        if setup is not None:
            params['setup'] = setup
        if legacy is not None:
            params['legacy'] = legacy

        data = {'source': source}
        if config is not None:
            data['configuration'] = config
        if dependencies is not None:
            data['dependencies'] = dependencies

        request = Request(
            method='post',
            endpoint='/_api/foxx',
            params=params,
            data=data,
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceCreateError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def create_service_with_file(
            self,
            mount,
            filename,
            development=None,
            setup=None,
            legacy=None):
        """Install a new service using a javascript file or zip bundle.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param filename: Full path to the javascript file or zip bundle.
        :type filename: str | unicode
        :param development: Enable development mode.
        :type development: bool
        :param setup: Run service setup script.
        :type setup: bool
        :param legacy: Install the service in 2.8 legacy compatibility mode.
        :type legacy: bool
        :return: Service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceCreateError: If install fails.
        """
        params = {'mount': mount}
        if development is not None:
            params['development'] = development
        if setup is not None:
            params['setup'] = setup
        if legacy is not None:
            params['legacy'] = legacy

        data = self._encode_file(filename)
        request = Request(
            method='post',
            endpoint='/_api/foxx',
            params=params,
            data=data,
            headers={'content-type': data.content_type}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceCreateError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def update_service(
            self,
            mount,
            source=None,
            config=None,
            dependencies=None,
            teardown=None,
            setup=None,
            legacy=None,
            force=None):
        """Update (upgrade) a service.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param source: Fully qualified URL or absolute path on the server file
            system. Must be accessible by the server, or by all servers if in
            a cluster.
        :type source: str | unicode
        :param config: Configuration values.
        :type config: dict
        :param dependencies: Dependency settings.
        :type dependencies: dict
        :param teardown: Run service teardown script.
        :type teardown: bool
        :param setup: Run service setup script.
        :type setup: bool
        :param legacy: Update the service in 2.8 legacy compatibility mode.
        :type legacy: bool
        :param force: Force update if no service is found.
        :type force: bool
        :return: Updated service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceUpdateError: If update fails.
        """
        params = {'mount': mount}
        if teardown is not None:
            params['teardown'] = teardown
        if setup is not None:
            params['setup'] = setup
        if legacy is not None:
            params['legacy'] = legacy
        if force is not None:
            params['force'] = force

        data = {}
        if source is not None:
            data['source'] = source
        if config is not None:
            data['configuration'] = config
        if dependencies is not None:
            data['dependencies'] = dependencies

        request = Request(
            method='patch',
            endpoint='/_api/foxx/service',
            params=params,
            data=data,
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceUpdateError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def update_service_with_file(
            self,
            mount,
            filename=None,
            teardown=None,
            setup=None,
            legacy=None,
            force=None):
        """Update (upgrade) a service using a javascript file or zip bundle.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param filename: Full path to the javascript file or zip bundle.
        :type filename: str | unicode
        :param teardown: Run service teardown script.
        :type teardown: bool
        :param setup: Run service setup script.
        :type setup: bool
        :param legacy: Update the service in 2.8 legacy compatibility mode.
        :type legacy: bool
        :param force: Force update if no service is found.
        :type force: bool
        :return: Updated service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceUpdateError: If update fails.
        """
        params = {'mount': mount}
        if teardown is not None:
            params['teardown'] = teardown
        if setup is not None:
            params['setup'] = setup
        if legacy is not None:
            params['legacy'] = legacy
        if force is not None:
            params['force'] = force

        data = self._encode_file(filename)
        request = Request(
            method='patch',
            endpoint='/_api/foxx/service',
            params=params,
            data=data,
            headers={'content-type': data.content_type}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceUpdateError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def replace_service(self,
                        mount,
                        source,
                        config=None,
                        dependencies=None,
                        teardown=None,
                        setup=None,
                        legacy=None,
                        force=None):
        """Replace a service by removing the old one and installing a new one.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param source: Fully qualified URL or absolute path on the server file
            system. Must be accessible by the server, or by all servers if in
            a cluster.
        :type source: str | unicode
        :param config: Configuration values.
        :type config: dict
        :param dependencies: Dependency settings.
        :type dependencies: dict
        :param teardown: Run service teardown script.
        :type teardown: bool
        :param setup: Run service setup script.
        :type setup: bool
        :param legacy: Replace the service in 2.8 legacy compatibility mode.
        :type legacy: bool
        :param force: Force install if no service is found.
        :type force: bool
        :return: Replaced service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceReplaceError: If replace fails.
        """
        params = {'mount': mount}
        if teardown is not None:
            params['teardown'] = teardown
        if setup is not None:
            params['setup'] = setup
        if legacy is not None:
            params['legacy'] = legacy
        if force is not None:
            params['force'] = force

        data = {}
        if source is not None:
            data['source'] = source
        if config is not None:
            data['configuration'] = config
        if dependencies is not None:
            data['dependencies'] = dependencies

        request = Request(
            method='put',
            endpoint='/_api/foxx/service',
            params=params,
            data=data,
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceReplaceError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def replace_service_with_file(
            self,
            mount,
            filename=None,
            teardown=None,
            setup=None,
            legacy=None,
            force=None):
        """Replace a service using a javascript file or zip bundle.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param filename: Full path to the javascript file or zip bundle.
        :type filename: str | unicode
        :param teardown: Run service teardown script.
        :type teardown: bool
        :param setup: Run service setup script.
        :type setup: bool
        :param legacy: Replace the service in 2.8 legacy compatibility mode.
        :type legacy: bool
        :param force: Force install if no service is found.
        :type force: bool
        :return: Replaced service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxServiceReplaceError: If replace fails.
        """
        params = {'mount': mount}
        if teardown is not None:
            params['teardown'] = teardown
        if setup is not None:
            params['setup'] = setup
        if legacy is not None:
            params['legacy'] = legacy
        if force is not None:
            params['force'] = force

        data = self._encode_file(filename)
        request = Request(
            method='put',
            endpoint='/_api/foxx/service',
            params=params,
            data=data,
            headers={'content-type': data.content_type}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceReplaceError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def delete_service(self, mount, teardown=None):
        """Uninstall a service.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param teardown: Run service teardown script.
        :type teardown: bool
        :return: True if service was deleted successfully.
        :rtype: bool
        :raise arango.exceptions.FoxxServiceDeleteError: If delete fails.
        """
        params = {'mount': mount}
        if teardown is not None:
            params['teardown'] = teardown

        request = Request(
            method='delete',
            endpoint='/_api/foxx/service',
            params=params
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxServiceDeleteError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def config(self, mount):
        """Return service configuration.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Configuration values.
        :rtype: dict
        :raise arango.exceptions.FoxxConfigGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx/configuration',
            params={'mount': mount},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxConfigGetError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def update_config(self, mount, config):
        """Update service configuration.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param config: Configuration values. Omitted options are ignored.
        :type config: dict
        :return: Updated configuration values.
        :rtype: dict
        :raise arango.exceptions.FoxxConfigUpdateError: If update fails.
        """
        request = Request(
            method='patch',
            endpoint='/_api/foxx/configuration',
            params={'mount': mount},
            data=config
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxConfigUpdateError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def replace_config(self, mount, config):
        """Replace service configuration.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param config: Configuration values. Omitted options are reset to their
            default values or marked as un-configured.
        :type config: dict
        :return: Replaced configuration values.
        :rtype: dict
        :raise arango.exceptions.FoxxConfigReplaceError: If replace fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/foxx/configuration',
            params={'mount': mount},
            data=config
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxConfigReplaceError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def dependencies(self, mount):
        """Return service dependencies.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Dependency settings.
        :rtype: dict
        :raise arango.exceptions.FoxxDependencyGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx/dependencies',
            params={'mount': mount},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxDependencyGetError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def update_dependencies(self, mount, dependencies):
        """Update service dependencies.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param dependencies: Dependencies settings. Omitted ones are ignored.
        :type dependencies: dict
        :return: Updated dependency settings.
        :rtype: dict
        :raise arango.exceptions.FoxxDependencyUpdateError: If update fails.
        """
        request = Request(
            method='patch',
            endpoint='/_api/foxx/dependencies',
            params={'mount': mount},
            data=dependencies
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxDependencyUpdateError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def replace_dependencies(self, mount, dependencies):
        """Replace service dependencies.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param dependencies: Dependencies settings. Omitted ones are disabled.
        :type dependencies: dict
        :return: Replaced dependency settings.
        :rtype: dict
        :raise arango.exceptions.FoxxDependencyReplaceError: If replace fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/foxx/dependencies',
            params={'mount': mount},
            data=dependencies
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxDependencyReplaceError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def enable_development(self, mount):
        """Put the service into development mode.

        While the service is running in development mode, it is reloaded from
        the file system, and its setup script (if any) is re-executed every
        time the service handles a request.

        In a cluster with multiple coordinators, changes to the filesystem on
        one coordinator is not reflected across other coordinators.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxDevModeEnableError: If operation fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/foxx/development',
            params={'mount': mount},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxDevModeEnableError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def disable_development(self, mount):
        """Put the service into production mode.

        In a cluster with multiple coordinators, the services on all other
        coordinators are replaced with the version on the calling coordinator.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Service metadata.
        :rtype: dict
        :raise arango.exceptions.FoxxDevModeDisableError: If operation fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/foxx/development',
            params={'mount': mount},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxDevModeDisableError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def readme(self, mount):
        """Return the service readme.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Service readme.
        :rtype: str | unicode
        :raise arango.exceptions.FoxxReadmeGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx/readme',
            params={'mount': mount},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxReadmeGetError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def swagger(self, mount):
        """Return the Swagger API description for the given service.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Swagger API description.
        :rtype: dict
        :raise arango.exceptions.FoxxSwaggerGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx/swagger',
            params={'mount': mount}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxSwaggerGetError(resp, request)
            if 'basePath' in resp.body:
                resp.body['base_path'] = resp.body.pop('basePath')
            return resp.body

        return await self._execute(request, response_handler)

    async def download(self, mount):
        """Download service bundle.

        When development mode is enabled, a new bundle is created every time.
        Otherwise, the bundle represents the version of the service installed
        on the server.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Service bundle in raw string form.
        :rtype: str | unicode
        :raise arango.exceptions.FoxxDownloadError: If download fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/foxx/download',
            params={'mount': mount}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxDownloadError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def commit(self, replace=None):
        """Commit local service state of the coordinator to the database.

        This can be used to resolve service conflicts between coordinators
        that cannot be fixed automatically due to missing data.

        :param replace: Overwrite any existing service files in database.
        :type replace: bool
        :return: True if the state was committed successfully.
        :rtype: bool
        :raise arango.exceptions.FoxxCommitError: If commit fails.
        """
        params = {}
        if replace is not None:
            params['replace'] = replace

        request = Request(
            method='post',
            endpoint='/_api/foxx/commit',
            params=params
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxCommitError(resp, request)
            return True

        return await self._execute(request, response_handler)

    async def scripts(self, mount):
        """List service scripts.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :return: Service scripts.
        :rtype: dict
        :raise arango.exceptions.FoxxScriptListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/foxx/scripts',
            params={'mount': mount},
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxScriptListError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def run_script(self, mount, name, arg=None):
        """Run a service script.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param name: Script name.
        :type name: str | unicode
        :param arg: Arbitrary value passed into the script as first argument.
        :type arg: str | unicode | bool | int | list | dict
        :return: Result of the script, if any.
        :rtype: dict
        :raise arango.exceptions.FoxxScriptRunError: If script fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/foxx/scripts/{}'.format(name),
            params={'mount': mount},
            data=arg or {}
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxScriptRunError(resp, request)
            return resp.body

        return await self._execute(request, response_handler)

    async def run_tests(
            self,
            mount,
            reporter='default',
            idiomatic=None,
            output_format=None,
            name_filter=None):
        """Run service tests.

        :param mount: Service mount path (e.g "/_admin/aardvark").
        :type mount: str | unicode
        :param reporter: Test reporter. Allowed values are "default" (simple
            list of test cases), "suite" (object of test cases nested in
            suites), "stream" (raw stream of test results), "xunit" (XUnit or
            JUnit compatible structure), or "tap" (raw TAP compatible stream).
        :type reporter: str | unicode
        :param idiomatic: Use matching format for the reporter, regardless of
            the value of parameter **output_format**.
        :type: bool
        :param output_format: Used to further control format. Allowed values
            are "x-ldjson", "xml" and "text". When using "stream" reporter,
            setting this to "x-ldjson" returns newline-delimited JSON stream.
            When using "tap" reporter, setting this to "text" returns plain
            text TAP report. When using "xunit" reporter, settings this to
            "xml" returns an XML instead of JSONML.
        :type output_format: str | unicode
        :param name_filter: Only run tests whose full name (test suite and
            test case) matches the given string.
        :type name_filter: str | unicode
        :return: Reporter output (e.g. raw JSON string, XML, plain text).
        :rtype: str | unicode
        :raise arango.exceptions.FoxxTestRunError: If test fails.
        """
        params = {'mount': mount, 'reporter': reporter}
        if idiomatic is not None:
            params['idiomatic'] = idiomatic
        if name_filter is not None:
            params['filter'] = name_filter

        headers = {}
        if output_format == 'x-ldjson':
            headers['Accept'] = 'application/x-ldjson'
        elif output_format == 'xml':
            headers['Accept'] = 'application/xml'
        elif output_format == 'text':
            headers['Accept'] = 'text/plain'

        request = Request(
            method='post',
            endpoint='/_api/foxx/tests',
            params=params,
            headers=headers
        )

        def response_handler(resp):
            if not resp.is_success:
                raise FoxxTestRunError(resp, request)
            return resp.raw_body

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

import pytest
from aioarangodb.exceptions import (
    IndexListError,
    IndexCreateError,
    IndexDeleteError,
    IndexLoadError
)
from aioarangodb.tests.helpers import assert_raises, extract
pytestmark = pytest.mark.asyncio


async def test_list_indexes(icol, bad_col):
    indexes = await icol.indexes()
    assert isinstance(indexes, list)
    assert len(indexes) > 0
    assert 'id' in indexes[0]
    assert 'type' in indexes[0]
    assert 'fields' in indexes[0]
    assert 'selectivity' in indexes[0]
    assert 'sparse' in indexes[0]
    assert 'unique' in indexes[0]

    with assert_raises(IndexListError) as err:
        await bad_col.indexes()
    assert err.value.error_code in {11, 1228}


async def test_add_hash_index(icol):
    icol = icol

    fields = ['attr1', 'attr2']
    result = await icol.add_hash_index(
        fields=fields,
        unique=True,
        sparse=True,
        deduplicate=True,
        name='hash_index',
        in_background=False
    )

    expected_index = {
        'sparse': True,
        'type': 'hash',
        'fields': ['attr1', 'attr2'],
        'unique': True,
        'deduplicate': True,
        'name': 'hash_index'
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Clean up the index
    await icol.delete_index(result['id'])


async def test_add_skiplist_index(icol):
    fields = ['attr1', 'attr2']
    result = await icol.add_skiplist_index(
        fields=fields,
        unique=True,
        sparse=True,
        deduplicate=True,
        name='skiplist_index',
        in_background=False
    )

    expected_index = {
        'sparse': True,
        'type': 'skiplist',
        'fields': ['attr1', 'attr2'],
        'unique': True,
        'deduplicate': True,
        'name': 'skiplist_index'
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Clean up the index
    await icol.delete_index(result['id'])


async def test_add_geo_index(icol):
    # Test add geo index with one attribute
    result = await icol.add_geo_index(
        fields=['attr1'],
        ordered=False,
        name='geo_index',
        in_background=True
    )

    expected_index = {
        'sparse': True,
        'type': 'geo',
        'fields': ['attr1'],
        'unique': False,
        'geo_json': False,
        'name': 'geo_index'
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Test add geo index with two attributes
    result = await icol.add_geo_index(
        fields=['attr1', 'attr2'],
        ordered=False,
    )
    expected_index = {
        'sparse': True,
        'type': 'geo',
        'fields': ['attr1', 'attr2'],
        'unique': False,
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Test add geo index with more than two attributes (should fail)
    with assert_raises(IndexCreateError) as err:
        await icol.add_geo_index(fields=['attr1', 'attr2', 'attr3'])
    assert err.value.error_code == 10

    # Clean up the index
    await icol.delete_index(result['id'])


async def test_add_fulltext_index(icol):
    # Test add fulltext index with one attributes
    result = await icol.add_fulltext_index(
        fields=['attr1'],
        min_length=10,
        name='fulltext_index',
        in_background=True
    )
    expected_index = {
        'sparse': True,
        'type': 'fulltext',
        'fields': ['attr1'],
        'min_length': 10,
        'unique': False,
        'name': 'fulltext_index'
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Test add fulltext index with two attributes (should fail)
    with assert_raises(IndexCreateError) as err:
        await icol.add_fulltext_index(fields=['attr1', 'attr2'])
    assert err.value.error_code == 10

    # Clean up the index
    await icol.delete_index(result['id'])


async def test_add_persistent_index(icol):
    # Test add persistent index with two attributes
    result = await icol.add_persistent_index(
        fields=['attr1', 'attr2'],
        unique=True,
        sparse=True,
        name='persistent_index',
        in_background=True
    )
    expected_index = {
        'sparse': True,
        'type': 'persistent',
        'fields': ['attr1', 'attr2'],
        'unique': True,
        'name': 'persistent_index'
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Clean up the index
    await icol.delete_index(result['id'])


async def test_add_ttl_index(icol):
    # Test add persistent index with two attributes
    result = await icol.add_ttl_index(
        fields=['attr1'],
        expiry_time=1000,
        name='ttl_index',
        in_background=True
    )
    expected_index = {
        'type': 'ttl',
        'fields': ['attr1'],
        'expiry_time': 1000,
        'name': 'ttl_index'
    }
    for key, value in expected_index.items():
        assert result[key] == value

    assert result['id'] in await extract('id', await icol.indexes())

    # Clean up the index
    await icol.delete_index(result['id'])


async def test_delete_index(icol, bad_col):
    old_indexes = set(await extract('id', await icol.indexes()))
    icol.add_hash_index(['attr3', 'attr4'], unique=True)
    icol.add_skiplist_index(['attr3', 'attr4'], unique=True)
    icol.add_fulltext_index(fields=['attr3'], min_length=10)

    new_indexes = set(await extract('id', await icol.indexes()))
    assert new_indexes.issuperset(old_indexes)

    indexes_to_delete = new_indexes - old_indexes
    for index_id in indexes_to_delete:
        assert await icol.delete_index(index_id) is True

    new_indexes = set(await extract('id', await icol.indexes()))
    assert new_indexes == old_indexes

    # Test delete missing indexes
    for index_id in indexes_to_delete:
        assert await icol.delete_index(index_id, ignore_missing=True) is False
    for index_id in indexes_to_delete:
        with assert_raises(IndexDeleteError) as err:
            await icol.delete_index(index_id, ignore_missing=False)
        assert err.value.error_code == 1212

    # Test delete indexes with bad collection
    for index_id in indexes_to_delete:
        with assert_raises(IndexDeleteError) as err:
            await bad_col.delete_index(index_id, ignore_missing=False)
        assert err.value.error_code in {11, 1228}


async def test_load_indexes(icol, bad_col):
    # Test load indexes
    assert await icol.load_indexes() is True

    # Test load indexes with bad collection
    with assert_raises(IndexLoadError) as err:
        await bad_col.load_indexes()
    assert err.value.error_code in {11, 1228}



from __future__ import absolute_import, unicode_literals, division
from aioarangodb.tests import global_data

import pytest
import asyncio
import os

from aioarangodb import ArangoClient
from aioarangodb.database import StandardDatabase
from aioarangodb import formatter

from aioarangodb.tests.arangodocker import arango_image
from aioarangodb.tests.helpers import (
    generate_db_name,
    generate_col_name,
    generate_string,
    generate_username,
    generate_graph_name,
    empty_collection,
    generate_jwt
)
from aioarangodb.tests.executors import (
    TestAsyncExecutor,
    TestBatchExecutor,
    TestTransactionExecutor
)


@pytest.fixture(scope='session')
def arango():
    if os.environ.get('TRAVIS', 'false') == 'false':
        host, port = arango_image.run()
    yield host, port
    if os.environ.get('TRAVIS', 'false') == 'false':
        arango_image.stop()

@pytest.fixture(scope='function')
async def client(arango):
    url = 'http://{}:{}'.format(
        'localhost',
        arango_image.get_port()
    )
    client = ArangoClient(hosts=[url, url, url])
    global_data['url'] = url
    global_data['client'] = client
    yield client
    await client.close()

@pytest.fixture(scope='function')
async def sys_db(client):
    sys_db = await client.db(
        name='_system',
        username='root',
        password=global_data.get('passwd'),
        superuser_token=generate_jwt(global_data.get('secret'))
    )
    username = generate_username()
    password = generate_string()

    global_data['username'] = username
    global_data['password'] = password
    await sys_db.create_database(
        name=global_data.get('tst_db_name'),
        users=[{
            'active': True,
            'username': username,
            'password': password,
        }]
    )
    yield sys_db
    await sys_db.clear_async_jobs()
    # Remove all test tasks.
    for task in await sys_db.tasks():
        task_name = task['name']
        if task_name.startswith('test_task'):
            await sys_db.delete_task(task_name, ignore_missing=True)

    # Remove all test users.
    for user in await sys_db.users():
        username = user['username']
        if username.startswith('test_user'):
            await sys_db.delete_user(username, ignore_missing=True)

    # Remove all test databases.
    for db_name in await sys_db.databases():
        if db_name.startswith('test_database'):
            await sys_db.delete_database(db_name, ignore_missing=True)

    # Remove all test collections.
    for collection in await sys_db.collections():
        col_name = collection['name']
        if col_name.startswith('test_collection'):
            await sys_db.delete_collection(col_name, ignore_missing=True)

@pytest.fixture(scope='function')
async def db(sys_db, client):
    tst_db = await client.db(global_data.get('tst_db_name'), global_data.get('username'), global_data.get('password'))

    # Create a standard collection for testing.
    col_name = generate_col_name()
    tst_col = await tst_db.create_collection(col_name, edge=False)
    await tst_col.add_skiplist_index(['val'])
    await tst_col.add_fulltext_index(['text'])
    geo_index = await tst_col.add_geo_index(['loc'])

    # Create a legacy edge collection for testing.
    icol_name = generate_col_name()
    await tst_db.create_collection(icol_name, edge=True)

    # Create test vertex & edge collections and graph.
    graph_name = generate_graph_name()
    ecol_name = generate_col_name()
    fvcol_name = generate_col_name()
    tvcol_name = generate_col_name()
    tst_graph = await tst_db.create_graph(graph_name)
    await tst_graph.create_vertex_collection(fvcol_name)
    await tst_graph.create_vertex_collection(tvcol_name)
    await tst_graph.create_edge_definition(
        edge_collection=ecol_name,
        from_vertex_collections=[fvcol_name],
        to_vertex_collections=[tvcol_name]
    )

    global_data.update({
        'geo_index': geo_index,
        'col_name': col_name,
        'icol_name': icol_name,
        'graph_name': graph_name,
        'ecol_name': ecol_name,
        'fvcol_name': fvcol_name,
        'tvcol_name': tvcol_name
    })

    return tst_db

@pytest.fixture(scope='function')
async def bad_db(sys_db, client):
    bad_db_name = generate_db_name()
    # Create a user and non-system database for testing.
    bad_db = await client.db(bad_db_name, global_data['username'], global_data['password'])
    return bad_db


@pytest.fixture(autouse=True)
def mock_formatters(monkeypatch):

    def mock_verify_format(body, result):
        body.pop('error', None)
        body.pop('code', None)
        result.pop('edge', None)
        if len(body) != len(result):
            raise ValueError(
                '\nIN: {}\nOUT: {}'.format(sorted(body), sorted(result))
            )
        return result

    monkeypatch.setattr(formatter, 'verify_format', mock_verify_format)

@pytest.fixture(autouse=False)
def url():
    return global_data['url']


@pytest.fixture(autouse=False)
def db_name():
    return global_data['tst_db_name']


@pytest.fixture(autouse=False)
def username():
    return global_data['username']


@pytest.fixture(autouse=False)
def password():
    return global_data['password']


@pytest.fixture(autouse=False)
def root_password():
    return global_data['passwd']


@pytest.fixture(autouse=False)
def conn(db):
    return getattr(db, '_conn')


@pytest.fixture(autouse=False)
async def col(db):
    collection = db.collection(global_data['col_name'])
    await empty_collection(collection)
    return collection


@pytest.fixture(autouse=False)
async def bad_col(bad_db):
    return bad_db.collection(global_data['col_name'])


@pytest.fixture(autouse=False)
def geo():
    return global_data['geo_index']


@pytest.fixture(autouse=False)
async def icol(db):
    collection = db.collection(global_data['icol_name'])
    await empty_collection(collection)
    return collection


@pytest.fixture(autouse=False)
async def graph(db):
    return db.graph(global_data['graph_name'])


@pytest.fixture(autouse=False)
async def bad_graph(bad_db):
    return bad_db.graph(global_data['graph_name'])


# noinspection PyShadowingNames
@pytest.fixture(autouse=False)
async def fvcol(graph):
    collection = graph.vertex_collection(global_data['fvcol_name'])
    await empty_collection(collection)
    return collection


# noinspection PyShadowingNames
@pytest.fixture(autouse=False)
async def tvcol(graph):
    collection = graph.vertex_collection(global_data['tvcol_name'])
    await empty_collection(collection)
    return collection


# noinspection PyShadowingNames
@pytest.fixture(autouse=False)
async def bad_fvcol(bad_graph):
    return bad_graph.vertex_collection(global_data['fvcol_name'])


# noinspection PyShadowingNames
@pytest.fixture(autouse=False)
async def ecol(graph):
    collection = graph.edge_collection(global_data['ecol_name'])
    await empty_collection(collection)
    return collection


# noinspection PyShadowingNames
@pytest.fixture(autouse=False)
async def bad_ecol(bad_graph):
    return bad_graph.edge_collection(global_data['ecol_name'])


@pytest.fixture(autouse=False)
def docs():
    return [
        {'_key': '1', 'val': 1, 'text': 'foo', 'loc': [1, 1]},
        {'_key': '2', 'val': 2, 'text': 'foo', 'loc': [2, 2]},
        {'_key': '3', 'val': 3, 'text': 'foo', 'loc': [3, 3]},
        {'_key': '4', 'val': 4, 'text': 'bar', 'loc': [4, 4]},
        {'_key': '5', 'val': 5, 'text': 'bar', 'loc': [5, 5]},
        {'_key': '6', 'val': 6, 'text': 'bar', 'loc': [5, 5]},
    ]


@pytest.fixture(autouse=False)
def fvdocs():
    return [
        {'_key': '1', 'val': 1},
        {'_key': '2', 'val': 2},
        {'_key': '3', 'val': 3},
    ]


@pytest.fixture(autouse=False)
def tvdocs():
    return [
        {'_key': '4', 'val': 4},
        {'_key': '5', 'val': 5},
        {'_key': '6', 'val': 6},
    ]


@pytest.fixture(autouse=False)
def edocs():
    fv = global_data['fvcol_name']
    tv = global_data['tvcol_name']
    return [
        {'_key': '1', '_from': '{}/1'.format(fv), '_to': '{}/4'.format(tv)},
        {'_key': '2', '_from': '{}/1'.format(fv), '_to': '{}/5'.format(tv)},
        {'_key': '3', '_from': '{}/6'.format(fv), '_to': '{}/2'.format(tv)},
        {'_key': '4', '_from': '{}/8'.format(fv), '_to': '{}/7'.format(tv)},
    ]


@pytest.fixture(autouse=False)
def cluster():
    return global_data['cluster']


@pytest.fixture(autouse=False)
def replication():
    return global_data['replication']


@pytest.fixture(autouse=False)
def secret():
    return global_data['secret']
